import asyncio
import signal
import time
import typing as tp

from croniter import croniter

from .base import Oxalis, Task, logger


class Beater:
    def __init__(self, oxalis: Oxalis) -> None:
        self.oxalis = oxalis
        self.tasks: tp.List[Task] = []
        self.croniteres: tp.List[croniter] = []
        self.crons: tp.List[str] = []
        self.futures: tp.List[asyncio.Future] = []
        self.running = False

    def register(self, cron: str, task: Task):
        self.tasks.append(task)
        self.crons.append(cron)
        self.croniteres.append(croniter(cron))

    async def beat(self, i: int):
        t = self.croniteres[i].get_next() - time.time()
        await asyncio.sleep(t)
        await self.tasks[i].delay()
        logger.info(f"Beat task {self.tasks[i]}")
        self.futures[i] = asyncio.ensure_future(self.beat(i))

    async def _run(self):
        while self.running:
            await asyncio.sleep(0.5)

    def close(self, *_):
        logger.info("Close beater...")
        self.running = False
        for f in self.futures:
            f.cancel()

    def run(self):
        self.running = True
        signal.signal(signal.SIGINT, self.close)
        signal.signal(signal.SIGTERM, self.close)
        asyncio.get_event_loop().run_until_complete(self.oxalis.connect())
        for i in range(len(self.tasks)):
            logger.info(f"Beat task: {self.tasks[i]} at <{self.crons[i]}> ...")
            self.futures.append(asyncio.ensure_future(self.beat(i)))
        asyncio.get_event_loop().run_until_complete(self._run())
        asyncio.get_event_loop().run_until_complete(self.oxalis.disconnect())



from __future__ import annotations

import asyncio
import typing as tp

import aio_pika

from .base import Oxalis as _Oxalis
from .base import Task as _Task
from .base import TaskCodec, logger
from .pool import Pool

ExchangeType = aio_pika.ExchangeType


class Exchange(aio_pika.Exchange):
    NAME_PREFIX = "oxalis_exchange_"

    def __init__(
        self,
        name: str,
        type: tp.Union[ExchangeType, str] = ExchangeType.DIRECT,
        *,
        auto_delete: bool = False,
        durable: bool = False,
        internal: bool = False,
        passive: bool = False,
        arguments: aio_pika.abc.Arguments = None,
    ):
        self._type = type.value if isinstance(type, ExchangeType) else type
        self.name = self.NAME_PREFIX + name
        self.auto_delete = auto_delete
        self.durable = durable
        self.internal = internal
        self.passive = passive
        self.arguments = arguments or {}

    def set_channel(self, channel: aio_pika.abc.AbstractChannel):
        self.channel = channel.channel


class Queue(aio_pika.Queue):
    NAME_PREFIX = "oxalis_queue_"

    def __init__(
        self,
        name: str,
        durable: bool = True,
        exclusive: bool = False,
        auto_delete: bool = False,
        arguments: tp.Optional[aio_pika.abc.Arguments] = None,
        passive: bool = False,
    ):
        self.__get_lock = asyncio.Lock()
        self.close_callbacks = aio_pika.tools.CallbackCollection(self)
        self.name = self.NAME_PREFIX + name
        self.durable = durable
        self.exclusive = exclusive
        self.auto_delete = auto_delete
        self.arguments = arguments
        self.passive = passive

    def set_channel(self, channel: aio_pika.abc.AbstractChannel):
        self.channel = channel.channel


class Task(_Task):
    def __init__(
        self,
        oxalis: Oxalis,
        func: tp.Callable,
        exchange: Exchange,
        routing_key: str,
        name="",
        timeout: float = -1,
        ack_later: bool = False,
    ) -> None:
        super().__init__(oxalis, func, name, timeout)
        self.exchange = exchange
        self.routing_key = routing_key
        self.ack_later = ack_later


class Oxalis(_Oxalis):
    channel: aio_pika.abc.AbstractChannel

    def __init__(
        self,
        connection: aio_pika.Connection,
        task_cls: tp.Type[Task] = Task,
        task_codec: TaskCodec = TaskCodec(),
        pool: Pool = Pool(),
        timeout: float = 5.0,
        worker_num: int = 0,
        test: bool = False,
        default_queue_name="default",
        default_exchange_name="default",
        default_routing_key="default",
    ) -> None:
        super().__init__(
            task_cls=task_cls,
            task_codec=task_codec,
            pool=pool,
            timeout=timeout,
            worker_num=worker_num,
            test=test,
        )
        self.connection = connection
        self.ack_later_tasks: tp.Set[str] = set()
        self.default_exchange = Exchange(default_exchange_name)
        self.default_queue = Queue(default_queue_name)
        self.default_routing_key = default_routing_key
        self.queues: tp.List[Queue] = [self.default_queue]
        self.exchanges: tp.List[Exchange] = [self.default_exchange]
        self.bindings: tp.List[tp.Tuple[Queue, Exchange, str]] = [
            (self.default_queue, self.default_exchange, self.default_routing_key)
        ]
        self.routing_keys: tp.Dict[str, str] = {}

    async def connect(self):
        await self.connection.connect(timeout=self.timeout)
        self.channel = self.connection.channel()
        await self.channel.initialize(timeout=self.timeout)
        await self.declare(self.queues)
        await self.declare(self.exchanges)
        for q, e, k in self.bindings:
            await self.bind(q, e, k)

    async def disconnect(self):
        await self.connection.close()

    async def send_task(self, task: Task, *task_args, **task_kwargs):  # type: ignore[override]
        if task.name not in self.tasks:
            raise ValueError(f"Task {task} not register")
        logger.debug(f"Send task {task} to worker...")
        task.exchange.set_channel(self.channel)
        await task.exchange.publish(
            aio_pika.Message(
                self.task_codec.encode(task, task_args, task_kwargs),
                content_type="text/plain",
            ),
            routing_key=task.routing_key,
            timeout=self.timeout,
        )

    def register(
        self,
        task_name: str = "",
        timeout: float = -1,
        exchange: tp.Optional[aio_pika.abc.AbstractExchange] = None,
        routing_key: str = "",
        ack_later: bool = False,
        **_,
    ) -> tp.Callable[[tp.Callable], Task]:
        def wrapped(func):
            task = self.task_cls(
                self,
                func,
                exchange or self.default_exchange,
                routing_key or self.default_routing_key,
                name=task_name,
                timeout=timeout,
                ack_later=ack_later,
            )
            self.register_task(task)
            self.exchanges.append(task.exchange)
            return task

        return wrapped

    def register_queues(self, queues: tp.Sequence[Queue]):
        self.queues.extend(queues)

    def register_exchanges(self, exchanges: tp.Sequence[Exchange]):
        self.exchanges.extend(exchanges)

    def register_binding(self, queue: Queue, exchange: Exchange, routing_key: str = ""):
        self.bindings.append((queue, exchange, routing_key))

    async def declare(self, eqs: tp.Sequence[tp.Union[Queue, Exchange]]):
        _names = set()
        for eq in eqs:
            if eq.name in _names:
                continue
            eq.set_channel(self.channel)
            await eq.declare(timeout=self.timeout)
            _names.add(eq.name)

    async def bind(self, queue: Queue, exchange: Exchange, routing_key: str = ""):
        queue.set_channel(self.channel)
        exchange.set_channel(self.channel)
        await queue.bind(exchange, routing_key, timeout=self.timeout)

    async def exec_task(self, task: Task, *args, **task_kwargs):  # type: ignore[override]
        message: aio_pika.IncomingMessage = args[0]
        task_args = args[1:]
        if not task.ack_later:
            await message.ack()
        await super().exec_task(task, *task_args, **task_kwargs)
        if task.ack_later:
            await message.ack()

    async def _receive_message(self, queue: Queue):
        async with self.connection.channel() as channel:
            while self.running:
                queue.set_channel(channel)
                try:
                    message = await queue.get(timeout=self.timeout)
                    if message:
                        await self.on_message_receive(message.body, message)
                except aio_pika.exceptions.QueueEmpty:
                    await asyncio.sleep(0.1)

    def _run_worker(self):
        queues = []
        _queues = set()
        for q in self.queues:
            if q.name in _queues:
                continue
            else:
                _queues.add(q.name)
                queues.append(q)

        for q in queues:
            asyncio.ensure_future(self._receive_message(q))



from __future__ import annotations

import abc
import asyncio
import inspect
import json
import logging
import multiprocessing
import os
import signal
import sys
import typing as tp

from .pool import Pool

logger = logging.getLogger("oxalis")


class Task:
    def __init__(
        self, oxalis: Oxalis, func: tp.Callable, name="", timeout: float = -1
    ) -> None:
        self.oxalis = oxalis
        self.func = func
        self.name = name or self.get_name()
        self.timeout = timeout

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}({self.name})>"

    async def __call__(self, *args: tp.Any, **kwargs: tp.Any) -> tp.Any:
        ret = self.func(*args, **kwargs)
        if inspect.iscoroutine(ret):
            ret = await ret

        return ret

    async def delay(self, *args, **kwargs) -> tp.Any:
        if self.oxalis.test:
            return await self(*args, **kwargs)
        else:
            await self.oxalis.send_task(self, *args, **kwargs)

    def get_name(self) -> str:
        return ".".join((self.func.__module__, self.func.__name__))


class TaskCodec:
    MESSAGE_TYPE = tp.Tuple[str, tp.Sequence[tp.Any], tp.Dict[str, tp.Any]]

    @classmethod
    def encode(
        cls,
        task: Task,
        task_args: tp.Sequence[tp.Any],
        task_kwargs: tp.Dict[str, tp.Any],
    ) -> bytes:
        return json.dumps([task.name, list(task_args), task_kwargs]).encode()

    @classmethod
    def decode(cls, content: bytes) -> MESSAGE_TYPE:
        return json.loads(content)


class Oxalis(abc.ABC):
    def __init__(
        self,
        task_cls: tp.Type[Task] = Task,
        task_codec: TaskCodec = TaskCodec(),
        pool: Pool = Pool(),
        timeout: float = 5.0,
        worker_num: int = 0,
        test: bool = False,
    ) -> None:
        self.task_cls = task_cls
        self.tasks: tp.Dict[str, Task] = {}
        self.task_codec = task_codec
        self.pool = pool
        self.running = False
        self.timeout = timeout
        self.test = test
        self._on_close_signal_count = 0
        self.worker_num = worker_num or os.cpu_count()

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}(pid-{os.getpid()})>"

    async def connect(self):
        pass

    async def disconnect(self):
        pass

    @abc.abstractmethod
    async def send_task(self, task: Task, *task_args, **task_kwargs):
        pass

    async def exec_task(self, task: Task, *task_args, **task_kwargs):
        logger.debug(f"Worker {self} execute task {task}...")
        await task(*task_args, **task_kwargs)

    def run_worker_master(self):
        signal.signal(signal.SIGINT, self.close)
        signal.signal(signal.SIGTERM, self.close)
        ps = []
        for _ in range(self.worker_num):
            ps.append(multiprocessing.Process(target=self.run_worker))
            ps[-1].start()
        for p in ps:
            p.join()

    def run_worker(self):
        logger.info(f"Run worker: {self}...")
        self.running = True
        self.on_worker_init()
        asyncio.get_event_loop().run_until_complete(self.connect())
        self._run_worker()
        asyncio.get_event_loop().run_until_complete(self.work())
        self.on_worker_close()

    @abc.abstractmethod
    def _run_worker(self):
        pass

    async def work(self):
        while self.running:
            await asyncio.sleep(self.timeout)
        await self.pool.close()
        await self.disconnect()

    def close_worker(self, force: bool = False):
        logger.info(f"Close worker{'(force)' if force else ''}: {self}...")
        self.running = False
        if force:
            sys.exit()

    def register_task(self, task: Task):
        if task.name in self.tasks:
            raise ValueError("double task, check task name")
        self.tasks[task.name] = task

    def register(
        self, task_name: str = "", timeout: float = -1, **_
    ) -> tp.Callable[[tp.Callable], Task]:
        def wrapped(func):
            task = self.task_cls(self, func, name=task_name, timeout=timeout)
            self.register_task(task)
            return task

        return wrapped

    async def on_message_receive(self, content: bytes, *args):
        try:
            task_name, task_args, task_kwargs = self.task_codec.decode(content)
        except Exception as e:
            logger.exception(e)
            return

        if task_name not in self.tasks:
            logger.warning(f"Received task {task_name} not found")
        else:
            await self.pool.block_spawn(
                self.exec_task(self.tasks[task_name], *args, *task_args, **task_kwargs),
                timeout=self.tasks[task_name].timeout,
            )

    def close(self, *_):
        self._on_close_signal_count += 1
        self.close_worker(force=self._on_close_signal_count >= 2)

    def on_worker_init(self):
        pass

    def on_worker_close(self):
        pass



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['oxalis']

package_data = \
{'': ['*']}

install_requires = \
['aio-pika>=8.0.3',
 'aioredis>=2.0.1',
 'async-timeout>=4.0.2',
 'croniter>=1.3.5']

setup_kwargs = {
    'name': 'oxalis',
    'version': '0.3.0',
    'description': 'Distributed async task/job queue',
    'long_description': '',
    'author': 'strongbugman',
    'author_email': 'strongbugman@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from __future__ import annotations

import asyncio
import typing as tp

from aioredis.client import Redis

from .base import Oxalis as _Oxalis
from .base import Task as _Task
from .base import TaskCodec, logger
from .pool import Pool


class Queue:
    NAME_PREFIX = "oxalis_queue_"

    def __init__(self, name: str) -> None:
        self.name = self.NAME_PREFIX + name


class PubsubQueue(Queue):
    pass


class Task(_Task):
    def __init__(
        self,
        oxalis: Oxalis,
        func: tp.Callable,
        queue: Queue,
        name="",
        timeout: float = -1,
    ) -> None:
        super().__init__(oxalis, func, name, timeout)
        self.queue = queue


class Oxalis(_Oxalis):
    def __init__(
        self,
        client: Redis,
        task_cls: tp.Type[Task] = Task,
        task_codec: TaskCodec = TaskCodec(),
        pool: Pool = Pool(),
        timeout: float = 5.0,
        worker_num: int = 0,
        test: bool = False,
        default_queue_name: str = "default",
    ) -> None:
        super().__init__(
            task_cls=task_cls,
            task_codec=task_codec,
            pool=pool,
            timeout=timeout,
            worker_num=worker_num,
            test=test,
        )
        self.client = client
        self.pubsub = client.pubsub()
        self.default_queue = Queue(default_queue_name)

    async def connect(self):
        await self.client.initialize()

    async def disconnect(self):
        await self.client.close()

    async def send_task(self, task: Task, *task_args, **task_kwargs):  # type: ignore[override]
        if task.name not in self.tasks:
            raise ValueError(f"Task {task} not register")
        logger.debug(f"Send task {task} to worker...")
        content = self.task_codec.encode(task, task_args, task_kwargs)
        if isinstance(task.queue, PubsubQueue):
            await self.client.publish(task.queue.name, content)
        else:
            await self.client.rpush(task.queue.name, content)

    def register(
        self,
        task_name: str = "",
        timeout: float = -1,
        queue: tp.Optional[Queue] = None,
        **_,
    ) -> tp.Callable[[tp.Callable], Task]:
        def wrapped(func):
            task = self.task_cls(
                self, func, queue or self.default_queue, name=task_name, timeout=timeout
            )
            self.register_task(task)
            return task

        return wrapped

    async def _receive_message(self):
        queues = list(
            {
                q.name
                for q in [t.queue for t in self.tasks.values()]
                if not isinstance(q, PubsubQueue)
            }
        )
        while self.running:
            content = await self.client.blpop(queues, timeout=self.timeout)
            if not content:
                continue
            else:
                await self.on_message_receive(content[1])

    async def _receive_pubsub_message(self):
        queues = {
            t.queue.name
            for t in self.tasks.values()
            if isinstance(t.queue, PubsubQueue)
        }
        while self.running:
            if not self.pubsub.subscribed:
                await self.pubsub.subscribe(*queues)
            content = await self.pubsub.get_message(
                ignore_subscribe_messages=True, timeout=self.timeout
            )
            if not content:
                continue
            else:
                await self.on_message_receive(content["data"])

    def _run_worker(self):
        if [t for t in self.tasks.values() if isinstance(t.queue, PubsubQueue)]:
            asyncio.ensure_future(self._receive_pubsub_message())
        if [t for t in self.tasks.values() if not isinstance(t.queue, PubsubQueue)]:
            asyncio.ensure_future(self._receive_message())



import asyncio
import logging
import typing as tp
from asyncio.queues import Queue, QueueEmpty

import async_timeout

logger = logging.getLogger("lotus")


class Pool:
    def __init__(self, limit: int = 100, timeout: tp.Union[int, float] = 5 * 60):
        self.limit = limit
        self.timeout = timeout
        self.pending_queue: Queue = Queue()
        self.done_queue: Queue = Queue()
        self.futures: tp.Set[asyncio.Future] = set()
        self.running_count = 0
        self.running = True

    def spawn(
        self, coroutine: tp.Awaitable, pending: bool = True, timeout: float = -1
    ) -> bool:
        if not self.running:
            raise RuntimeError("This pool has been closed")

        timeout = timeout if timeout >= 0 else self.timeout
        while True:
            if self.limit == -1 or self.running_count < self.limit:
                self.running_count += 1
                f = asyncio.ensure_future(self.run_coroutine(coroutine, timeout))
                f.add_done_callback(self.on_future_done)
                self.futures.add(f)
                break
            elif pending:
                self.pending_queue.put_nowait(coroutine)
                break
            else:
                return False

        return True

    async def block_spawn(self, coroutine: tp.Awaitable, timeout: float = -1) -> bool:
        while True:
            if self.spawn(coroutine, pending=False, timeout=timeout):
                break
            else:
                await self.done_queue.get()

        return True

    async def run_coroutine(self, coroutine: tp.Awaitable, timeout: float = -1):
        async with async_timeout.timeout(timeout):
            await coroutine

    @property
    def done(self) -> bool:
        return not (self.running_count or self.pending_queue.qsize())

    async def wait_done(self):
        while not self.done:
            await self.done_queue.get()

    async def close(self, force=False):
        self.running = False
        if force:
            for f in self.futures:
                f.cancel()
        await self.wait_done()

    def check_future(self, f: asyncio.Future):
        e = f.exception()
        if e:
            try:
                raise e
            except Exception:
                logger.exception(e)

    def on_future_done(self, f: asyncio.Future):
        self.running_count -= 1
        self.done_queue.put_nowait(f)
        self.check_future(f)
        self.futures.remove(f)
        try:
            if self.running and (self.limit == -1 or self.running_count < self.limit):
                next = self.pending_queue.get_nowait()
                self.running_count += 1
                asyncio.ensure_future(next).add_done_callback(self.on_future_done)
        except QueueEmpty:
            pass
